/* lpcboot - an LPC17xx MMC/SD boot loader inspired by HolgerBootloader2
   Based on a heavily modified version of ChaN's FatFs library

   Copyright (C) 2013  Ingo Korb <ingo@akana.de>
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.


   config.h: configuration options

*/

#ifndef CONFIG_H
#define CONFIG_H

#include <arm/NXP/LPC17xx/LPC17xx.h>
#include <arm/bits.h>
#include "autoconf.h"

/* ----- Common definitions for all LPC17xx hardware variants ------ */

/* PLL settings for 100MHz from a 12MHz crystal */
#define PLL_MULTIPLIER 25
#define PLL_PREDIV     2
#define PLL_DIVISOR    3

/* SSP clock divisors (must be even) - 400kHz slow, 16.7MHz fast */
#define SSP_CLK_DIVISOR_FAST 6
#define SSP_CLK_DIVISOR_SLOW 250

/* flash size and offsets */
#define FLASH_SIZE         (512 * 1024)
#define BOOTINFO_OFFSET    256
#define APPLICATION_OFFSET 16384
#define APPLICATION_LENGTH (FLASH_SIZE - APPLICATION_OFFSET)

#if CONFIG_HARDWARE_VARIANT == example
/* ---------- Hardware configuration: Example ---------- */
/* This is a commented example for most of the available options    */
/* in case someone wants to build Yet Another[tm] hardware variant. */
/* Some of the values are chosen randomly, so this variant is not   */
/* expected to compile successfully.                                */

/*** SD card support ***/
/* SD Card supply voltage - choose the one appropiate to your board */
/* #  define SD_SUPPLY_VOLTAGE (1L<<15)  / * 2.7V - 2.8V */
/* #  define SD_SUPPLY_VOLTAGE (1L<<16)  / * 2.8V - 2.9V */
/* #  define SD_SUPPLY_VOLTAGE (1L<<17)  / * 2.9V - 3.0V */
#  define SD_SUPPLY_VOLTAGE (1L<<18)  /* 3.0V - 3.1V */
/* #  define SD_SUPPLY_VOLTAGE (1L<<19)  / * 3.1V - 3.2V */
/* #  define SD_SUPPLY_VOLTAGE (1L<<20)  / * 3.2V - 3.3V */
/* #  define SD_SUPPLY_VOLTAGE (1L<<21)  / * 3.3V - 3.4V */
/* #  define SD_SUPPLY_VOLTAGE (1L<<22)  / * 3.4V - 3.5V */
/* #  define SD_SUPPLY_VOLTAGE (1L<<23)  / * 3.5V - 3.6V */

#  define SPI_ON_SSP        1
#  define SD_CS_PIN         6


/*** LEDs ***/

/* Initialize ports for all LEDs */
static inline void leds_init(void) {
  /* 22 onboard red, 27 red, 28 yellow */
  LPC_GPIO0->FIODIR |= BV(22) | BV(27) | BV(28);
  LPC_GPIO0->FIOSET  =          BV(27) | BV(28);
  LPC_GPIO0->FIOCLR  = BV(22);
  /* green */
  LPC_GPIO2->FIODIR |= BV(13);
  LPC_GPIO2->FIOSET  = BV(13);
}

/* reset all GPIOs to their power-up default state */
static inline void gpio_deinit(void) {
  LPC_GPIO0->FIODIR   = 0;
  LPC_GPIO2->FIODIR   = 0;
  LPC_GPIO0->FIOCLR   = 0xffffffff;
  LPC_GPIO2->FIOCLR   = 0xffffffff;
  LPC_PINCON->PINSEL0 = 0;
}

static inline void sdcard_interface_init(void) {
  /* configure SD0-CS as output, high */
  LPC_GPIO0->FIOSET  = BV(SD_CS_PIN);
  LPC_GPIO0->FIODIR |= BV(SD_CS_PIN);

  /* Connect SSP1 */
  LPC_PINCON->PINSEL0 |= BV(15) | BV(17) | BV(19); // SCK, MISO, MOSI
}

static inline __attribute__((always_inline)) void set_green_led(uint8_t state) {
  if (state)
    BITBAND(LPC_GPIO2->FIOCLR, 13) = 1;
  else
    BITBAND(LPC_GPIO2->FIOSET, 13) = 1;
}

static inline __attribute__((always_inline)) void set_red_led(uint8_t state) {
  if (state)
    BITBAND(LPC_GPIO0->FIOCLR, 27) = 1;
  else
    BITBAND(LPC_GPIO0->FIOSET, 27) = 1;
}

static inline __attribute__((always_inline)) void set_test_led(uint8_t state) {
  /* onboard red */
  if (state)
    BITBAND(LPC_GPIO0->FIOSET, 22) = 1;
  else
    BITBAND(LPC_GPIO0->FIOCLR, 22) = 1;
}

/*** Bootloader info ***/
/* Default bootloader device ID if CONFIG_BOOT_DEVID is not set */
#define BOOTLOADER_DEVID 0xdeadbeef


/* Pre-configurated hardware variants */
#elif CONFIG_HARDWARE_VARIANT == 101
/* ---------- Hardware configuration: arm2iec v1.x ---------- */
#  define SD_SUPPLY_VOLTAGE (1L<<21)  /* 3.3V - 3.4V */
#  define SPI_ON_SSP        1
#  define SD_CS_PIN         6

static inline void sdcard_set_ss(int state) {
  if (state)
    LPC_GPIO0->FIOSET = BV(SD_CS_PIN);
  else
    LPC_GPIO0->FIOCLR = BV(SD_CS_PIN);
}

static inline void leds_init(void) {
  /* 22 onboard red, 27 red, 28 yellow */
  LPC_GPIO0->FIODIR |= BV(22) | BV(27) | BV(28);
  LPC_GPIO0->FIOSET  =          BV(27) | BV(28);
  LPC_GPIO0->FIOCLR  = BV(22);
  /* green */
  LPC_GPIO2->FIODIR |= BV(13);
  LPC_GPIO2->FIOSET  = BV(13);
}

static inline void gpio_deinit(void) {
  LPC_GPIO0->FIODIR = 0;
  LPC_GPIO2->FIODIR = 0;
  LPC_GPIO0->FIOCLR = 0xffffffff;
  LPC_GPIO2->FIOCLR = 0xffffffff;
  LPC_PINCON->PINSEL0 = 0;
}

static inline void sdcard_interface_init(void) {
  /* configure SD0-CS as output, high */
  LPC_GPIO0->FIOSET  = BV(SD_CS_PIN);
  LPC_GPIO0->FIODIR |= BV(SD_CS_PIN);

  /* Connect SSP1 */
  LPC_PINCON->PINSEL0 |= BV(15) | BV(17) | BV(19); // SCK, MISO, MOSI
}

static inline __attribute__((always_inline)) void set_green_led(uint8_t state) {
  if (state)
    BITBAND(LPC_GPIO2->FIOCLR, 13) = 1;
  else
    BITBAND(LPC_GPIO2->FIOSET, 13) = 1;
}

static inline __attribute__((always_inline)) void set_red_led(uint8_t state) {
  if (state)
    BITBAND(LPC_GPIO0->FIOCLR, 27) = 1;
  else
    BITBAND(LPC_GPIO0->FIOSET, 27) = 1;
}

#  define BOOTLOADER_DEVID 0x31693261

#else
#  error "CONFIG_HARDWARE_VARIANT is unset or set to an unknown value."
#endif

/* ---------------- End of user-configurable options ---------------- */

#ifdef CONFIG_BOOT_DEVID
#  undef  BOOTLOADER_DEVID
#  define BOOTLOADER_DEVID CONFIG_BOOT_DEVID
#endif

#define CPUCLK_KHZ         (CONFIG_MCU_FREQ / 1000)

#endif
