/* lpcboot - an LPC17xx MMC/SD boot loader inspired by HolgerBootloader2
   Based on a heavily modified version of ChaN's FatFs library

   Copyright (C) 2013  Ingo Korb <ingo@akana.de>
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.


   lpc_iap.c: Interface to the in-application programming ROM

*/

#include <arm/NXP/LPC17xx/LPC17xx.h>
#include <arm/bits.h>
#include "config.h"
#include "lpc_iap.h"

#define IAP_ADDRESS 0x1fff1ff1

/* command codes */
#define IAP_PREPARE      50
#define IAP_RAM2FLASH    51
#define IAP_ERASE        52
#define IAP_BLANK_CHECK  53
#define IAP_READ_PARTID  54
#define IAP_READ_BOOTVER 55
#define IAP_COMPARE      56
#define IAP_RUN_ISP      57
#define IAP_READ_SERIAL  58

unsigned char sectorsizes[LPC_SECTOR_COUNT+1] = {
   4,  4,  4,  4, // 16K boot loader
   4,  4,  4,  4,
   4,  4,  4,  4,
   4,  4,  4,  4,
  32, 32, 32, 32,
  32, 32, 32, 32,
  32, 32, 32, 32,
  32, 32, 0
};

unsigned int iap_address_to_sector(void *address) {
  unsigned int i, curaddr;
  unsigned int intaddr = (unsigned int)address;

  intaddr /= 1024;
  curaddr = 0;
  i = 0;

  do {
    curaddr += sectorsizes[i];
    i++;
  } while (curaddr <= intaddr && sectorsizes[i] != 0);

  return i-1;
}

static unsigned long command[5];
static unsigned long result[5];

typedef void (*iapfunc_t)(unsigned long command[], unsigned long result[]);
static const iapfunc_t IAP = (iapfunc_t)IAP_ADDRESS;

iapresult_t iap_read_partid(unsigned long *id) {
  command[0] = IAP_READ_PARTID;
  IAP(command, result);
  *id = result[1];
  return result[0];
}

iapresult_t iap_read_serial(unsigned long serial[const 4]) {
  command[0] = IAP_READ_SERIAL;
  IAP(command, result);
  serial[0] = result[1];
  serial[1] = result[2];
  serial[2] = result[3];
  serial[3] = result[4];
  return result[0];
}

iapresult_t iap_read_bootcode_version(unsigned long *version) {
  command[0] = IAP_READ_BOOTVER;
  IAP(command, result);
  *version = result[1];
  return result[0];
}

iapresult_t iap_blank_check(unsigned int startsector, unsigned int endsector) {
  command[0] = IAP_BLANK_CHECK;
  command[1] = startsector;
  command[2] = endsector;
  IAP(command, result);
  return result[0];
}

iapresult_t iap_prepare(unsigned int startsector, unsigned int endsector) {
  command[0] = IAP_PREPARE;
  command[1] = startsector;
  command[2] = endsector;
  IAP(command, result);
  return result[0];
}

iapresult_t iap_erase(unsigned int startsector, unsigned int endsector, unsigned int cpu_khz) {
  command[0] = IAP_ERASE;
  command[1] = startsector;
  command[2] = endsector;
  command[3] = cpu_khz;
  IAP(command, result);
  return result[0];
}

iapresult_t iap_ram2flash(void *dest, void *src, unsigned int bytes, unsigned int cpu_khz) {
  command[0] = IAP_RAM2FLASH;
  command[1] = (unsigned long)dest;
  command[2] = (unsigned long)src;
  command[3] = bytes;
  command[4] = cpu_khz;
  IAP(command, result);
  return result[0];
}

iapresult_t iap_compare(void *addr1, void *addr2, unsigned int bytes) {
  command[0] = IAP_COMPARE;
  command[1] = (unsigned long)addr1;
  command[2] = (unsigned long)addr2;
  command[3] = bytes;
  IAP(command, result);
  return result[0];
}

void iap_start_isp(void) {
  asm volatile("cpsid i");
  command[0] = IAP_RUN_ISP;
  IAP(command, result);
}
